/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Linker.h"

void Linker::addFromParser(const Parser &p) {
    for (auto &fn : p.getTree()) {
        addFunction(fn);
    }
    for (auto &v : p.getVariables()) {
        variables[v.first] = v.second;
    }
    for (auto &v : p.getReservations()) {
        reservations[v.first] = v.second;
    }
}

void Linker::addFunction(const Function &f) {
    for (auto &fn : functions) {
        if (fn.label == f.label) {
            fn = f;
            return;
        }
    }
    functions.push_back(f);
}

void Linker::addVariable(const std::string &k, const std::vector<Token> &v) {
    variables[k] = v;
}

std::string Linker::getTextConstant(const std::string &name, const std::string &defaultValue, bool throwOnNotFound) const {
    if(getVariables().find(name) == getVariables().end()){
        if(throwOnNotFound){
            throw std::runtime_error(name + " is not defined");
        }
        return defaultValue;
    }
    for(auto& t : getVariables().at(name)){
        return t.getTextValue();
    }
    if(throwOnNotFound){
        throw std::runtime_error(name + " is not set properly");
    }
    return defaultValue;
}

uint32_t Linker::getIntegerConstant(const std::string &name, uint32_t defaultValue, bool throwOnNotFound) const {
    if(getVariables().find(name) == getVariables().end()){
        if(throwOnNotFound){
            throw std::runtime_error(name + " is not defined");
        }
        return defaultValue;
    }
    for(auto& t : getVariables().at(name)){
        if(t.getType() == Token::Type::Immediate || t.getType() == Token::Type::RegisterLocation || t.getType() == Token::Type::CodeLocation){
            return t.getNumericValue();
        }
    }
    if(throwOnNotFound){
        throw std::runtime_error(name + " is not set properly");
    }
    return defaultValue;
}

void Linker::addReservation(uint32_t a, const std::string &v) {
    if(getReservations().find(a) != getReservations().end()){
        throw std::runtime_error(v + " already reserved as " + getReservations().at(a));
    }
    reservations[a] = v;
}
